@@ -16,8 +16,7 @@ from .utils.itertools import cartesian_product
 class KbNufftModule(KbModule):
     """Parent class for KbNufft classes.
 
-    This implementation collects all init functions into one place. It inherits
-    from torch.nn.Module via torchkbnufft.kbmodule.KbModule.
+    This implementation collects all init functions into one place.
 
     Args:
         im_size (int or tuple of ints): Size of base image.
@@ -38,11 +37,12 @@ class KbNufftModule(KbModule):
 
     def __init__(self, im_size, grid_size=None, numpoints=6, n_shift=None,
                  table_oversamp=2**10, kbwidth=2.34, order=0, norm='None',
-                 coil_broadcast=False, matadj=False):
+                 coil_broadcast=False, matadj=False, grad_traj=False):
         super(KbNufftModule, self).__init__()
 
         self.im_size = im_size
         self.im_rank = len(im_size)
+        self.grad_traj = grad_traj
         if grid_size is None:
             self.grid_size = tuple(np.array(self.im_size) * 2)
         else:
@@ -136,6 +136,7 @@ class KbNufftModule(KbModule):
         interpob['norm'] = self.norm
         interpob['coil_broadcast'] = self.coil_broadcast
         interpob['matadj'] = self.matadj
+        interpob['grad_traj'] = self.grad_traj
         Jgen = []
         for i in range(self.im_rank):
             # number of points to use for interpolation is numpoints
@@ -145,13 +146,13 @@ class KbNufftModule(KbModule):
 
         return interpob
 
-def kbnufft_forward(interpob):
+def kbnufft_forward(interpob, multiprocessing=False):
     @tf.function(experimental_relax_shapes=True)
     @tf.custom_gradient
     def kbnufft_forward_for_interpob(x, om):
         """Apply FFT and interpolate from gridded data to scattered data.
 
-        Inputs are assumed to be batch/chans x coil x real/imag x image dims.
+        Inputs are assumed to be batch/chans x coil x image dims.
         Om should be nbatch x ndims x klength.
 
         Args:
@@ -169,29 +170,40 @@ def kbnufft_forward(interpob):
         grid_size = interpob['grid_size']
         im_size = interpob['im_size']
         norm = interpob['norm']
+        grad_traj = interpob['grad_traj']
         im_rank = interpob.get('im_rank', 2)
 
-        x = scale_and_fft_on_image_volume(
-            x, scaling_coef, grid_size, im_size, norm, im_rank=im_rank)
+        fft_x = scale_and_fft_on_image_volume(
+            x, scaling_coef, grid_size, im_size, norm, im_rank=im_rank, multiprocessing=multiprocessing)
 
-        y = kbinterp(x, om, interpob)
+        y = kbinterp(fft_x, om, interpob)
 
         def grad(dy):
-            x = adjkbinterp(dy, om, interpob)
-            x = ifft_and_scale_on_gridded_data(
-                x, scaling_coef, grid_size, im_size, norm, im_rank=im_rank)
-            return x, None
+            # Gradients with respect to image
+            grid_dy = adjkbinterp(dy, om, interpob)
+            ifft_dy = ifft_and_scale_on_gridded_data(
+                grid_dy, scaling_coef, grid_size, im_size, norm, im_rank=im_rank)
+            if grad_traj:
+                # Gradients with respect to trajectory locations
+                r = [tf.linspace(-im_size[i]/2, im_size[i]/2-1, im_size[i]) for i in range(im_rank)]
+                grid_r = tf.cast(tf.meshgrid(*r, indexing='ij'), x.dtype)[None, ...]
+                fft_dx_dom = scale_and_fft_on_image_volume(
+                x * grid_r, scaling_coef, grid_size, im_size, norm, im_rank=im_rank)
+                dy_dom = tf.cast(-1j * dy * kbinterp(fft_dx_dom, om, interpob), tf.float32)
+            else:
+                dy_dom = None
+            return ifft_dy, dy_dom
 
         return y, grad
     return kbnufft_forward_for_interpob
 
-def kbnufft_adjoint(interpob):
+def kbnufft_adjoint(interpob, multiprocessing=False):
     @tf.function(experimental_relax_shapes=True)
     @tf.custom_gradient
     def kbnufft_adjoint_for_interpob(y, om):
         """Interpolate from scattered data to gridded data and then iFFT.
 
-        Inputs are assumed to be batch/chans x coil x real/imag x kspace
+        Inputs are assumed to be batch/chans x coil x kspace
         length. Om should be nbatch x ndims x klength.
 
         Args:
@@ -201,25 +213,32 @@ def kbnufft_adjoint(interpob):
         Returns:
             tensor: The image after adjoint NUFFT.
         """
-        x = adjkbinterp(y, om, interpob)
-
+        grid_y = adjkbinterp(y, om, interpob)
         scaling_coef = interpob['scaling_coef']
         grid_size = interpob['grid_size']
         im_size = interpob['im_size']
         norm = interpob['norm']
+        grad_traj = interpob['grad_traj']
         im_rank = interpob.get('im_rank', 2)
-
-        x = ifft_and_scale_on_gridded_data(
-            x, scaling_coef, grid_size, im_size, norm, im_rank=im_rank)
+        ifft_y = ifft_and_scale_on_gridded_data(
+            grid_y, scaling_coef, grid_size, im_size, norm, im_rank=im_rank, multiprocessing=multiprocessing)
 
         def grad(dx):
-            x = scale_and_fft_on_image_volume(
+            # Gradients with respect to off grid signal
+            fft_dx = scale_and_fft_on_image_volume(
                 dx, scaling_coef, grid_size, im_size, norm, im_rank=im_rank)
-
-            y = kbinterp(x, om, interpob)
-
-            return y, None
-        return x, grad
+            dx_dy = kbinterp(fft_dx, om, interpob)
+            if grad_traj:
+                # Gradients with respect to trajectory locations
+                r = [tf.linspace(-im_size[i]/2, im_size[i]/2-1, im_size[i]) for i in range(im_rank)]
+                grid_r = tf.cast(tf.meshgrid(*r, indexing='ij'), dx.dtype)[None, ...]
+                ifft_dxr = scale_and_fft_on_image_volume(
+                    dx * grid_r, scaling_coef, grid_size, im_size, norm, im_rank=im_rank, do_ifft=True)
+                dx_dom = tf.cast(1j * y * kbinterp(ifft_dxr, om, interpob, conj=True), om.dtype)
+            else:
+                dx_dom = None
+            return dx_dy, dx_dom
+        return ifft_y, grad
     return kbnufft_adjoint_for_interpob
 
 
